"""YAGDSPY - Yet Another Graph Dependancy System in PYthon. Provide
core functions for building processing pipelines, whether for building
like make/maven/scons, or multi-step analysis using files to save
intermediate or final results.
"""
# Copyright 2016, 2017 Fred Hutchinson Cancer Research Center

# This file is to tell setuptools that this directory is a package

from .processing_module import *

